/* dosfsck.c  -  User interface */

/* Written 1993 by Werner Almesberger */

/* FAT32, VFAT, Atari format support, and various fixes additions May 1998
 * by Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de> */


#if defined(__DJGPP__) || defined(__rtems__)
#include "version.h"
#if defined(__DJGPP__)
#include <crt0.h>
int _crt0_startup_flags = _CRT0_FLAG_LOCK_MEMORY | _CRT0_FLAG_DROP_EXE_SUFFIX;
    /* disable all swapping / VM, and suppress the ".exe" in argv[0] string */
    /* also nice: _CRT0_FLAG_FILL_SBRK_MEMORY _CRT0_FLAG_FILL_DEADBEEF */
#endif
#else
#include "../version.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#ifndef __DJGPP__
#if defined(__rtems__)
#define __need_getopt_newlib
#endif
#include <getopt.h>	/* in DJGPP there is some getopt in unistd */
#endif

#include "common.h"
#include "dosfsck.h"
#include "io.h"
#include "boot.h"
#include "fat.h"
#include "file.h"
#include "check.h"

int interactive = 0,list = 0,test = 0,verbose = 0,write_immed = 0;
int atari_format = 0;
unsigned n_files = 0;
void *mem_queue = NULL;

#if defined(__rtems__)

#include <setjmp.h>

static jmp_buf dosfsck_exit_jmp;
static int dosfsck_exit_code;

void dosfsck_exit(int code)
{
  dosfsck_exit_code = code;
  longjmp (dosfsck_exit_jmp, 1);
}

static int dosfsck_main(int argc,char **argv);

int
rtems_shell_main_dosfsck(int argc, char *argv[])
{
  interactive = 0,list = 0,test = 0,verbose = 0,write_immed = 0;
  atari_format = 0;
  n_files = 0;
  mem_queue = NULL;
  if (setjmp (dosfsck_exit_jmp) == 0)
    return dosfsck_main (argc, argv);
  return dosfsck_exit_code;
}

#define main dosfsck_main
#define basename(a) "msdosfsck"

#endif

static void usage(char *name)
{
    fprintf(stderr,"usage: %s [-aAflrtvVwy] [-d path -d ...] "
      "[-u path -u ...]\n%15sdevice\n",name,"");
    fprintf(stderr,"  -a       automatically repair the file system\n");
    fprintf(stderr,"  -A       toggle Atari file system format\n");
    fprintf(stderr,"  -d path  drop that file\n");
    fprintf(stderr,"  -f       salvage unused chains to files\n");
    fprintf(stderr,"  -l       list path names\n");
    fprintf(stderr,"  -n       no-op, check non-interactively without changing\n");
    fprintf(stderr,"  -r       interactively repair the file system\n");
    fprintf(stderr,"  -t       test for bad clusters\n");
    fprintf(stderr,"  -u path  try to undelete that (non-directory) file\n");
    fprintf(stderr,"  -v       verbose mode\n");
    fprintf(stderr,"  -V       perform a verification pass\n");
    fprintf(stderr,"  -w       write changes to disk immediately\n");
    fprintf(stderr,"  -y       same as -a, for compat with other *fsck\n");
    exit(2); /* not die(...) */
}


/*
 * ++roman: On m68k, check if this is an Atari; if yes, turn on Atari variant
 * of MS-DOS filesystem by default.
 */
static void check_atari( void )
{
#if defined(__mc68000__) && !defined(__rtems__)
    FILE *f;
    char line[128], *p;

    if (!(f = fopen( "/proc/hardware", "r" ))) {
	perror( "/proc/hardware" );
	return;
    }

    while( fgets( line, sizeof(line), f ) ) {
	if (strncmp( line, "Model:", 6 ) == 0) {
	    p = line + 6;
	    p += strspn( p, " \t" );
	    if (strncmp( p, "Atari ", 6 ) == 0)
		atari_format = 1;
	    break;
	}
    }
    fclose( f );
#endif
}


int main(int argc,char **argv)
{
    DOS_FS fs;
    int rw,salvage_files,verify,c;
    unsigned long free_clusters;

#if defined(__rtems__)
    struct getopt_data getopt_reent;
    memset(&getopt_reent, 0, sizeof(getopt_data));
#define GETOPT_REENT(o) getopt_reent.o
#else
#define GETOPT_REENT(o) o
#endif
    
    rw = salvage_files = verify = 0;
    interactive = 1;
    check_atari();

#if defined(__rtems__)
    while ((c = getopt_r(argc,argv,"Aad:flnrtu:vVwy", &getopt_reent)) != EOF)
#else
    while ((c = getopt(argc,argv,"Aad:flnrtu:vVwy")) != EOF)
#endif
	switch (c) {
	    case 'A': /* toggle Atari format */
	  	atari_format = !atari_format;
		break;
	    case 'a':
	    case 'y':
		rw = 1;
		interactive = 0;
		salvage_files = 1;
		break;
	    case 'd':
		file_add(optarg,fdt_drop);
		break;
	    case 'f':
		salvage_files = 1;
		break;
	    case 'l':
		list = 1;
		break;
            case 'n':    
                rw = 0;  
                interactive = 0;
                break;
	    case 'r':
		rw = 1;
		interactive = 1;
		break;
	    case 't':
		test = 1;
		break;
	    case 'u':
		file_add(optarg,fdt_undelete);
		break;
	    case 'v': /* VERSION number is printed below anyway, v is VERBOSE */
		verbose = 1;
		/* printf("dosfsck " VERSION " (" VERSION_DATE ")\n"); */
		break;
	    case 'V':
		verify = 1;
		break;
	    case 'w':
		write_immed = 1;
		break;
	    default:
		usage(basename(argv[0]));
	}
    if ((test || write_immed) && !rw) {
	fprintf(stderr,"-t and -w require -a or -r\n");
	exit(2); /* not die(...) */
    }
    if (GETOPT_REENT(optind) != argc-1) usage(basename(argv[0]));

    printf( "dosfsck " VERSION ", " VERSION_DATE ", FAT32, LFN\n" );
    fs_open(argv[GETOPT_REENT(optind)],rw);
    read_boot(&fs);
    if (verify) printf("Starting check/repair pass.\n");
    while (read_fat(&fs), scan_root(&fs)) qfree(&mem_queue);
    if (test) fix_bad(&fs);
    if (salvage_files) reclaim_file(&fs);
    else reclaim_free(&fs);
    free_clusters = update_free(&fs);
    file_unused();
    qfree(&mem_queue);
    if (verify) {
	printf("Starting verification pass.\n");
	read_fat(&fs);
	scan_root(&fs);
	reclaim_free(&fs);
	qfree(&mem_queue);
    }

    if (fs_changed()) {
	if (rw) {
	    if (interactive)
		rw = get_key("yn","Perform changes ? (y/n)") == 'y';
	    else printf("Performing changes.\n");
	}
	else
	    printf("Leaving file system unchanged.\n");
    }

    printf( "%s: %u files, %lu/%lu clusters\n", argv[optind],
	    n_files, fs.clusters - free_clusters, fs.clusters );

    showalloc(); /* 2.11b */
    return fs_close(rw) ? 1 : 0;
}

/* Local Variables: */
/* tab-width: 8     */
/* End:             */
